from immutabledict import immutabledict
import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class EventAreaPerChannel(strax.Plugin):
    """Simple plugin that provides area per channel for main and alternative S1/S2 in the event."""

    depends_on = ("event_basics", "peaks")
    provides = ("event_area_per_channel", "event_n_channel")
    data_kind = immutabledict(zip(provides, ("events", "events")))
    __version__ = "0.1.1"

    compressor = "zstd"

    n_top_pmts = straxen.URLConfig(default=straxen.n_top_pmts, type=int, help="Number of top PMTs")

    def infer_dtype(self):
        # setting data type from peak dtype
        pfields_ = self.deps["peaks"].dtype_for("peaks").fields
        # populating data type
        infoline = {
            "s1": "main S1",
            "s2": "main S2",
            "alt_s1": "alternative S1",
            "alt_s2": "alternative S2",
        }
        dtype = []
        # populating APC
        ptypes = ["s1", "s2", "alt_s1", "alt_s2"]
        for type_ in ptypes:
            dtype += [
                (
                    (f"Area per channel for {infoline[type_]}", f"{type_}_area_per_channel"),
                    pfields_["area_per_channel"][0],
                )
            ]
            dtype += [
                (
                    (f"Length of the interval in samples for {infoline[type_]}", f"{type_}_length"),
                    pfields_["length"][0],
                )
            ]
            dtype += [
                (
                    (f"Width of one sample for {infoline[type_]} [ns]", f"{type_}_dt"),
                    pfields_["dt"][0],
                )
            ]
        # populating S1 n channel properties
        n_channel_dtype = [
            (("Main S1 count of contributing PMTs", "s1_n_channels"), np.int16),
            (("Main S1 top count of contributing PMTs", "s1_top_n_channels"), np.int16),
            (("Main S1 bottom count of contributing PMTs", "s1_bottom_n_channels"), np.int16),
        ]
        return {
            "event_area_per_channel": dtype + n_channel_dtype + strax.time_fields,
            "event_n_channel": n_channel_dtype + strax.time_fields,
        }

    def compute(self, events, peaks):
        area_per_channel = np.zeros(len(events), self.dtype["event_area_per_channel"])
        area_per_channel["time"] = events["time"]
        area_per_channel["endtime"] = strax.endtime(events)
        n_channel = np.zeros(len(events), self.dtype["event_n_channel"])
        n_channel["time"] = events["time"]
        n_channel["endtime"] = strax.endtime(events)

        split_peaks = strax.split_by_containment(peaks, events)
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            for type_ in ["s1", "s2", "alt_s1", "alt_s2"]:
                type_index = event[f"{type_}_index"]
                if type_index != -1:
                    type_area_per_channel = sp["area_per_channel"][type_index]
                    area_per_channel[f"{type_}_area_per_channel"][event_i] = type_area_per_channel
                    area_per_channel[f"{type_}_length"][event_i] = sp["length"][type_index]
                    area_per_channel[f"{type_}_dt"][event_i] = sp["dt"][type_index]
                    if type_ == "s1":
                        area_per_channel["s1_n_channels"][event_i] = (
                            type_area_per_channel > 0
                        ).sum()
                        area_per_channel["s1_top_n_channels"][event_i] = (
                            type_area_per_channel[: self.n_top_pmts] > 0
                        ).sum()
                        area_per_channel["s1_bottom_n_channels"][event_i] = (
                            type_area_per_channel[self.n_top_pmts :] > 0
                        ).sum()
        for field in ["s1_n_channels", "s1_top_n_channels", "s1_bottom_n_channels"]:
            n_channel[field] = area_per_channel[field]
        result = {
            "event_area_per_channel": area_per_channel,
            "event_n_channel": n_channel,
        }
        return result



import numpy as np
import numba
from scipy.stats import halfcauchy
import strax
import straxen

from .peak_ambience import distance_in_xy, _quick_assign

export, __all__ = strax.exporter()


@export
class PeakShadow(strax.OverlapWindowPlugin):
    """This plugin can find and calculate the time & position shadow from previous peaks in time. It
    also gives the area and (x,y) of the previous peaks.

    References:
        * v0.1.5 reference: xenon:xenonnt:ac:prediction:shadow_ambience

    """

    __version__ = "0.1.6"

    depends_on = ("peak_basics", "peak_positions")
    provides = "peak_shadow"
    save_when = strax.SaveWhen.EXPLICIT

    shadow_time_window_backward = straxen.URLConfig(
        default=int(1e9),
        type=int,
        track=True,
        help="Search for peaks casting time & position shadow in this time window [ns]",
    )

    shadow_threshold = straxen.URLConfig(
        default={"s1_time_shadow": 1e3, "s2_time_shadow": 1e4, "s2_position_shadow": 1e4},
        type=dict,
        track=True,
        help="Only take S1/S2s larger than this into account when calculating Shadow [PE]",
    )

    shadow_deltatime_exponent = straxen.URLConfig(
        default=-1.0, type=float, track=True, help="The exponent of delta t when calculating shadow"
    )

    shadow_sigma_and_baseline = straxen.URLConfig(
        default=[15.220, 0.036],
        type=list,
        track=True,
        help=(
            "Fitted position correlation sigma[cm*PE^0.5] and baseline[cm] using in position shadow"
        ),
    )

    def get_window_size(self):
        return (10 * self.shadow_time_window_backward, 0)

    def infer_dtype(self):
        s1_time_shadow_dtype = []
        s2_time_shadow_dtype = []
        s2_position_shadow_dtype = []
        nearest_dtype = []
        # We have time shadow(S2/dt) and position shadow(S2/dt*p(s))
        # previous S1 can only cast time shadow, previous S2 can cast both time & position shadow
        for key, dtype in zip(
            ["s1_time_shadow", "s2_time_shadow", "s2_position_shadow"],
            [s1_time_shadow_dtype, s2_time_shadow_dtype, s2_position_shadow_dtype],
        ):
            type_str, tp_desc, _ = key.split("_")
            dtype.append(
                (
                    (
                        f"previous large {type_str} casted largest {tp_desc} shadow [PE/ns]",
                        f"shadow_{key}",
                    ),
                    np.float32,
                )
            )
            dtype.append(
                (
                    (
                        (
                            f"time difference to the previous large {type_str} peak casting largest"
                            f" {tp_desc} shadow [ns]"
                        ),
                        f"dt_{key}",
                    ),
                    np.int64,
                )
            )
            # Only previous S2 peaks have (x,y)
            if "s2" in key:
                dtype.append(
                    (
                        (
                            f"x of previous large s2 peak casting largest {tp_desc} shadow [cm]",
                            f"x_{key}",
                        ),
                        np.float32,
                    )
                )
                dtype.append(
                    (
                        (
                            f"y of previous large s2 peak casting largest {tp_desc} shadow [cm]",
                            f"y_{key}",
                        ),
                        np.float32,
                    )
                )
            # Only time shadow gives the nearest large peak
            if "time" in key:
                dtype.append(
                    (
                        (
                            f"time difference to the nearest previous large {type_str} [ns]",
                            f"nearest_dt_{type_str}",
                        ),
                        np.int64,
                    )
                )
        # Also record the PDF of HalfCauchy when calculating S2 position shadow
        s2_position_shadow_dtype.append(
            (
                ("PDF describing correlation to the previous large s2", "pdf_s2_position_shadow"),
                np.float32,
            )
        )

        dtype = (
            s1_time_shadow_dtype
            + s2_time_shadow_dtype
            + s2_position_shadow_dtype
            + nearest_dtype
            + strax.time_fields
        )
        return dtype

    @property
    def shadowdtype(self):
        dtype = []
        dtype += [("shadow", np.float32), ("dt", np.int64)]
        dtype += [("x", np.float32), ("y", np.float32)]
        dtype += [("nearest_dt", np.int64)]
        return dtype

    def compute(self, peaks):
        argsort = strax.stable_argsort(peaks["center_time"])
        _peaks = strax.stable_sort(peaks, order="center_time")
        result = np.zeros(len(peaks), self.dtype)
        _quick_assign(argsort, result, self.compute_shadow(peaks, _peaks))
        return result

    def compute_shadow(self, peaks, current_peak):
        # 1. Define time window for each peak, we will find previous peaks within these time windows
        roi_shadow = np.zeros(len(current_peak), dtype=strax.time_fields)
        roi_shadow["time"] = current_peak["center_time"] - self.shadow_time_window_backward
        roi_shadow["endtime"] = current_peak["center_time"]

        # 2. Calculate S2 position shadow, S2 time shadow, and S1 time shadow
        result = np.zeros(len(current_peak), self.dtype)
        for key in ["s2_position_shadow", "s2_time_shadow", "s1_time_shadow"]:
            is_position = "position" in key
            type_str = key.split("_")[0]
            stype = 2 if "s2" in key else 1
            mask_pre = (peaks["type"] == stype) & (peaks["area"] > self.shadow_threshold[key])
            split_peaks = strax.touching_windows(peaks[mask_pre], roi_shadow)
            array = np.zeros(len(current_peak), np.dtype(self.shadowdtype))

            # Initialization
            array["x"] = np.nan
            array["y"] = np.nan
            array["dt"] = self.shadow_time_window_backward
            # The default value for shadow is set to be the lowest possible value
            if "time" in key:
                array["shadow"] = (
                    self.shadow_threshold[key] * array["dt"] ** self.shadow_deltatime_exponent
                )
            else:
                array["shadow"] = 0
            array["nearest_dt"] = self.shadow_time_window_backward

            # Calculating shadow, the Major of the plugin.
            # Only record the previous peak casting the largest shadow
            if len(current_peak):
                self.peaks_shadow(
                    current_peak,
                    peaks[mask_pre],
                    split_peaks,
                    self.shadow_deltatime_exponent,
                    array,
                    is_position,
                    self.getsigma(self.shadow_sigma_and_baseline, current_peak["area"]),
                )

            # Fill results
            names = ["shadow", "dt"]
            if "s2" in key:  # Only previous S2 peaks have (x,y)
                names += ["x", "y"]
            if "time" in key:  # Only time shadow gives the nearest large peak
                names += ["nearest_dt"]
            for name in names:
                if name == "nearest_dt":
                    result[f"{name}_{type_str}"] = array[name]
                else:
                    result[f"{name}_{key}"] = array[name]

        distance = np.sqrt(
            (result[f"x_s2_position_shadow"] - current_peak["x"]) ** 2
            + (result[f"y_s2_position_shadow"] - current_peak["y"]) ** 2
        )
        # If distance is NaN, set largest distance
        distance = np.where(np.isnan(distance), 2 * straxen.tpc_r, distance)
        # HalfCauchy PDF when calculating S2 position shadow
        result["pdf_s2_position_shadow"] = halfcauchy.pdf(
            distance, scale=self.getsigma(self.shadow_sigma_and_baseline, current_peak["area"])
        )

        # 6. Set time and endtime for peaks
        result["time"] = current_peak["time"]
        result["endtime"] = strax.endtime(current_peak)
        return result

    @staticmethod
    @np.errstate(invalid="ignore")
    def getsigma(sigma_and_baseline, s2):
        # The parameter of HalfCauchy, which is a function of S2 area
        return sigma_and_baseline[0] / np.sqrt(s2) + sigma_and_baseline[1]

    @staticmethod
    @numba.njit
    def peaks_shadow(peaks, pre_peaks, touching_windows, exponent, result, pos_corr, sigmas=None):
        """For each peak in peaks, check if there is a shadow-casting peak and check if it casts the
        largest shadow."""
        for p_i, (suspicious_peak, sigma) in enumerate(zip(peaks, sigmas)):
            # casting_peak is the previous large peak casting shadow
            # suspicious_peak is the suspicious peak which in shadow from casting_peak
            indices = touching_windows[p_i]
            for idx in range(indices[0], indices[1]):
                casting_peak = pre_peaks[idx]
                dt = suspicious_peak["center_time"] - casting_peak["center_time"]
                if dt <= 0:
                    continue
                # First we record the time difference to the nearest previous peak
                result["nearest_dt"][p_i] = min(result["nearest_dt"][p_i], dt)
                # Calculate time shadow
                new_shadow = casting_peak["area"] * dt**exponent
                if pos_corr:
                    # Calculate position shadow which is
                    # time shadow with a HalfCauchy PDF multiplier
                    distance = distance_in_xy(suspicious_peak, casting_peak)
                    distance = np.where(np.isnan(distance), 2 * straxen.tpc_r, distance)
                    new_shadow *= 2 / (np.pi * sigma * (1 + (distance / sigma) ** 2))
                # Only the previous peak with largest shadow is recorded
                if new_shadow > result["shadow"][p_i]:
                    result["shadow"][p_i] = new_shadow
                    result["x"][p_i] = casting_peak["x"]
                    result["y"][p_i] = casting_peak["y"]
                    result["dt"][p_i] = suspicious_peak["center_time"] - casting_peak["center_time"]



# Entrypoint loading logic
# Can be used to register context methods, mini-analyses,
# URLConfig protocols etc by other packeges when straxen is
# imported.
# The nice optimization of the different import attempts
# was copied from the hypothesis package
# https://github.com/HypothesisWorks/hypothesis/blob/master/hypothesis-python/src/hypothesis/entry_points.py


try:
    # We prefer to use importlib.metadata, or the backport on Python <= 3.7,
    # because it's much faster than pkg_resources (200ms import time speedup).
    try:
        from importlib import metadata as importlib_metadata
    except ImportError:
        import importlib_metadata  # type: ignore  # mypy thinks this is a redefinition

    def get_entry_points():
        try:
            eps = importlib_metadata.entry_points(group="straxen")
        except TypeError:
            # Load-time selection requires Python >= 3.10 or importlib_metadata >= 3.6,
            # so we'll retain this fallback logic for some time to come. See also
            # https://importlib-metadata.readthedocs.io/en/latest/using.html
            eps = importlib_metadata.entry_points().get("straxen", [])
        yield from eps

except ImportError:
    # But if we're not on Python >= 3.8 and the importlib_metadata backport
    # is not installed, we fall back to pkg_resources anyway.
    try:
        import pkg_resources
    except ImportError:
        import warnings

        warnings.warn(
            "Under Python <= 3.7, straxen requires either the importlib_metadata "
            "or setuptools package in order to load plugins via entrypoints.",
        )

        def get_entry_points():
            yield from ()

    else:

        def get_entry_points():
            yield from pkg_resources.iter_entry_points("straxen")


def load_entry_points():
    for entry in get_entry_points():  # pragma: no cover
        hook = entry.load()
        if callable(hook):
            hook()



from typing import Tuple, Dict, Union
import numpy as np
from tqdm import tqdm
from scipy.stats import norm, poisson
import numba
import strax
import straxen
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO, FAR_XYPOS_S2_TYPE, WIDE_XYPOS_S2_TYPE
from straxen.plugins.peaklets.peaklets import drop_data_field


export, __all__ = strax.exporter()


@export
class MergedS2s(strax.OverlapWindowPlugin):
    """Merge together peaklets if peak finding favours that they would form a single peak instead.

    Reference: xenon:xenonnt:analysis:s2_merging_time_position

    """

    __version__ = "2.0.0"

    depends_on: Tuple[str, ...] = (
        "peaklets",
        f"peaklet_positions_{DEFAULT_POSREC_ALGO}",
        "peaklet_classification",
        "lone_hits",
    )
    provides: Union[Tuple[str, ...], str] = ("merged_s2s", "enhanced_peaklet_classification")
    data_kind: Union[Dict[str, str], str] = dict(
        merged_s2s="merged_s2s", enhanced_peaklet_classification="peaklets"
    )

    n_tpc_pmts = straxen.URLConfig(type=int, help="Number of TPC PMTs")

    n_top_pmts = straxen.URLConfig(type=int, help="Number of top TPC array PMTs")

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )

    s2_merge_max_duration = straxen.URLConfig(
        default=50_000,
        type=int,
        infer_type=False,
        help="Do not merge peaklets at all if the result would be a peak longer than this [ns]",
    )

    s2_merge_gap_thresholds = straxen.URLConfig(
        default=(
            (1.84, 2.84e04),
            (2.18, 2.40e04),
            (2.51, 1.96e04),
            (2.84, 1.80e04),
            (3.18, 1.68e04),
            (3.51, 1.86e04),
            (3.84, 1.98e04),
            (4.18, 1.66e04),
            (4.51, 1.21e04),
        ),
        infer_type=False,
        help=(
            "Points to define maximum separation between peaklets to allow "
            "merging [ns] depending on log10 area of the merged peak\n"
            "where the gap size of the first point is the maximum gap to allow merging."
            "The format is ((log10(area), max_gap), (..., ...), (..., ...))"
        ),
    )

    s2_merge_p_thresholds = straxen.URLConfig(
        default=(
            (4.18, 9.89e-04),
            (4.51, 7.15e-03),
            (4.84, 1.60e-01),
        ),
        infer_type=False,
        help=(
            "Points to define minimum p-value of merging proposal "
            "depending on log10 area of the merged peak\n"
            "The format is ((log10(area), p-value), (..., ...), (..., ...))"
        ),
    )

    s2_merge_dr_thresholds = straxen.URLConfig(
        default=(
            (1.51, 1.40e01),
            (1.84, 1.83e01),
            (2.18, 1.97e01),
            (2.51, 1.24e01),
            (2.84, 5.75e00),
        ),
        type=tuple,
        help=(
            "Points to define maximum weighted mean deviation of "
            "the peaklets from the main cluster [cm]\n"
            "The format is ((log10(area_top), dr), (..., ...), (..., ...))"
        ),
    )

    s2_merge_unmerged_thresholds = straxen.URLConfig(
        default=(1, 49.5, 0.01),
        type=tuple,
        help=(
            "Max (number, fraction of unmerged, total area of unmerged) "
            "of type 20 peaklets inside a peak. "
            "The number of type 20 peaklets should not be larger than the number threshold. "
            "The area of type 20 peaklets should not be larger than the both area thresholds. "
            "The fraction threshold is important when S2 is large, "
            "while the total area threshold is important when S2 is small."
        ),
    )

    merge_lone_hits = straxen.URLConfig(
        default=True,
        type=bool,
        help="Merge lone hits into merged S2s",
    )

    merge_s0 = straxen.URLConfig(
        default=True,
        type=bool,
        help="Merge S0s into merged S2s",
    )

    gain_model = straxen.URLConfig(
        infer_type=False,
        help="PMT gain model. Specify as (str(model_config), str(version), nT-->boolean",
    )

    rough_seg = straxen.URLConfig(
        default=30, type=(int, float), help="Rough single electron gain [PE/e]"
    )

    sigma_seg = straxen.URLConfig(
        default=6.5,
        type=(int, float),
        help="Standard deviation of the single electron gain [PE/e]",
    )

    rough_min_sigma = straxen.URLConfig(
        default=1e2, type=(int, float), help="Minimum sigma for the merged peaks"
    )

    rough_max_sigma = straxen.URLConfig(
        default=2e4, type=(int, float), help="Maximum sigma for the merged peaks"
    )

    rough_sigma_bins = straxen.URLConfig(
        default=10, type=int, help="Number of bins for sigma of merged peaks"
    )

    rough_mu_bins = straxen.URLConfig(
        default=10, type=int, help="Number of bins for mu of merged peaks"
    )

    poisson_max_mu = straxen.URLConfig(
        default=25, type=int, help="When to switch from Poisson to normal distribution"
    )

    poisson_survival_ratio = straxen.URLConfig(
        default=1e-4,
        type=float,
        help=(
            "Survival ratio for Poisson distribution. The PMF smaller than this will be ignored."
        ),
    )

    normal_max_sigma = straxen.URLConfig(
        default=7,
        type=(int, float),
        help="Maximum sigma for the normal distribution CDF panel",
    )

    normal_n_bins = straxen.URLConfig(
        default=501,
        type=int,
        help="Number of bins for the normal distribution CDF panel",
    )

    maxexp = straxen.URLConfig(
        default=10,
        type=(int, float),
        help="Maximum exponent for the posterior to keep numerical stability",
    )

    use_bayesian_merging = straxen.URLConfig(default=True, type=bool, help="Use Bayesian merging")

    rm_sparse_xy = straxen.URLConfig(
        default=True, type=bool, help="Remove peaklets that are too far away in (x, y)"
    )

    use_uncertainty_weights = straxen.URLConfig(
        default=True, type=bool, help="Use uncertainty from probabilistic posrec to derive weights"
    )

    p_value_prioritized = straxen.URLConfig(
        default=False,
        type=bool,
        help="Whether to prioritize p-value over area when testing the proposal",
    )

    merged_s2s_get_window_size_factor = straxen.URLConfig(
        default=5, type=int, track=False, help="Factor of the window size for the merged_s2s plugin"
    )

    disable_progress_bar = straxen.URLConfig(
        default=True, type=bool, track=False, help="Whether to disable the progress bar"
    )

    copied_dtype = np.dtype(
        [
            ("time", np.int64),
            ("endtime", np.int64),
            ("area", np.float32),
            ("median_time", np.float32),
            ("area_decile_from_midpoint", np.float32, (11,)),
        ]
    )

    def _have_data(self, field):
        return field in self.deps["peaklets"].dtype_for("peaklets").names

    def infer_dtype(self):
        peaklet_classification_dtype = self.deps["peaklet_classification"].dtype_for(
            "peaklet_classification"
        )
        peaklets_dtype = self.deps["peaklets"].dtype_for("peaklets")
        # The merged dtype is argument position dependent!
        # It must be first classification then peaklet
        # Otherwise strax will raise an error
        # when checking for the returned dtype!
        merged_s2s_dtype = strax.merged_dtype((peaklet_classification_dtype, peaklets_dtype))
        enhanced_peaklet_classification_dtype = self.deps["peaklet_classification"].dtype_for(
            "peaklet_classification"
        )
        return dict(
            merged_s2s=merged_s2s_dtype,
            enhanced_peaklet_classification=enhanced_peaklet_classification_dtype,
        )

    def setup(self):
        self.to_pe = self.gain_model
        self.gap_thresholds = np.array(self.s2_merge_gap_thresholds).T
        self.p_thresholds = np.array(self.s2_merge_p_thresholds).T
        self.dr_thresholds = np.array(self.s2_merge_dr_thresholds).T
        # Max gap and area should be set by the gap thresholds to avoid contradictions
        if np.argmax(self.gap_thresholds[1]) != 0:
            raise ValueError("The first point should be the maximum gap to allow merging")
        if self.p_thresholds[1].max() > 1 or self.p_thresholds[1].min() < 0:
            raise ValueError("P-value should be in the range of [0, 1]")
        self.max_gap = self.gap_thresholds[1, 0]
        self.max_duration = self.s2_merge_max_duration

        self.poisson_max_k = np.ceil(
            poisson.isf(q=self.poisson_survival_ratio, mu=self.poisson_max_mu)
        ).astype(int)
        self.factorial_panel = np.array(
            [np.prod(np.arange(1, k + 1, dtype=float)) for k in range(self.poisson_max_k + 1)]
        )
        if np.any(self.factorial_panel < 0):
            raise ValueError("Factorial panel has negative values, this might because of overflow")
        x = np.linspace(-self.normal_max_sigma, self.normal_max_sigma, self.normal_n_bins)
        self.normal_panel = np.vstack([x, norm.cdf(x)])

        self.sigma = np.linspace(self.rough_min_sigma, self.rough_max_sigma, self.rough_sigma_bins)
        self.sigma_panel = np.repeat(self.sigma[None, :], self.rough_mu_bins, axis=0).flatten()

        self.unmerged_thresholds = self.s2_merge_unmerged_thresholds

    def get_window_size(self):
        return self.merged_s2s_get_window_size_factor * (
            int(self.s2_merge_gap_thresholds[0][1]) + self.s2_merge_max_duration
        )

    def no_merging(self, peaklets):
        is_s2 = peaklets["type"] == 2
        return np.sum(is_s2) <= 1 or self.max_gap < 0

    def compute(self, peaklets, lone_hits, start, end):
        if self.use_uncertainty_weights:
            name = f"position_contour_{self.default_reconstruction_algorithm}"
            if name not in peaklets.dtype.names:
                raise ValueError(f"{name} is not in the input peaklets dtype")

        # initialize enhanced_peaklet_classification
        enhanced_peaklet_classification = np.zeros(
            len(peaklets), dtype=self.dtype_for("enhanced_peaklet_classification")
        )
        # copy fields, especially type
        for d in enhanced_peaklet_classification.dtype.names:
            enhanced_peaklet_classification[d] = peaklets[d]

        if self.no_merging(peaklets):
            empty_result = self.empty_result()
            empty_result["enhanced_peaklet_classification"] = enhanced_peaklet_classification
            return empty_result

        # make sure the peaklets are not overwritten
        peaklets.flags.writeable = False

        is_s2 = peaklets["type"] == 2

        # peaklets might be overwritten in the merge method
        # so do not reuse the peaklets after this line
        merged_s2s, merged = self.merge(peaklets, lone_hits, start, end)

        # mark the peaklets can be merged by time-density but not
        # by position-density as type FAR_XYPOS_S2_TYPE
        enhanced_peaklet_classification["type"][is_s2 & ~merged] = FAR_XYPOS_S2_TYPE

        return dict(
            merged_s2s=merged_s2s, enhanced_peaklet_classification=enhanced_peaklet_classification
        )

    def merge(self, _peaklets, lone_hits, start, end):
        """Merge into S2s if the peaklets are close enough in time and position."""
        # this is an OverlapWindowPlugin, some peaklets will be reused in the next iteration
        # use _peaklets here to prevent peaklets from being overwritten
        if np.any(_peaklets["time"][1:] < strax.endtime(_peaklets)[:-1]):
            raise ValueError("Peaklets not disjoint, why?")

        # only keep S2 peaklets for merging
        is_s2 = _peaklets["type"] == 2

        if not (self._have_data("data_top") and self._have_data("data_start")):
            peaklets = np.zeros(
                is_s2.sum(),
                dtype=strax.peak_dtype(
                    n_channels=self.n_tpc_pmts, store_data_top=True, store_data_start=True
                ),
            )
            strax.copy_to_buffer(_peaklets[is_s2], peaklets, "_add_data_top_or_start_field")
        else:
            peaklets = _peaklets[is_s2]

        # make sure the peaklets are not overwritten
        peaklets.flags.writeable = False

        max_buffer = int(self.max_duration // strax.gcd_of_array(peaklets["dt"]))

        start_merge_at, end_merge_at, _merged = self.get_merge_instructions(
            peaklets,
            start,
            end,
            self.max_gap,
            self.max_duration,
            self.sigma,
            self.rough_seg,
            self.sigma_seg,
            self.rough_mu_bins,
            self.poisson_max_mu,
            self.poisson_survival_ratio,
            self.normal_panel,
            self.factorial_panel,
            self.sigma_panel,
            self.maxexp,
            self.n_top_pmts,
            self.p_thresholds,
            self.dr_thresholds,
            self.default_reconstruction_algorithm,
            self.use_bayesian_merging,
            self.rm_sparse_xy,
            self.use_uncertainty_weights,
            self.p_value_prioritized,
            self.gap_thresholds,
            disable=self.disable_progress_bar,
        )

        if "data_top" not in peaklets.dtype.names or "data_start" not in peaklets.dtype.names:
            raise ValueError("data_top or data_start is not in the peaklets dtype")

        # have to redo the merging to prevent numerical instability
        is_s0 = _peaklets["type"] == 0
        if self.merge_s0 and len(start_merge_at) and is_s0.sum() > 0:
            # build the time interval of merged_s2s, even though they are not merged yet
            endtime = strax.endtime(peaklets)
            merged_s2s_window = np.zeros(len(start_merge_at), dtype=strax.time_fields)
            for i in range(len(start_merge_at)):
                sl = slice(start_merge_at[i], end_merge_at[i])
                merged_s2s_window["time"][i] = peaklets["time"][sl][_merged[sl]][0]
                merged_s2s_window["endtime"][i] = endtime[sl][_merged[sl]][-1]
            # the S0s that should be merged should fully be contained
            merged_s0s = strax.split_by_containment(_peaklets[is_s0], merged_s2s_window)
            # offsets of indices
            increments = np.array([len(m) for m in merged_s0s], dtype=int)
            offsets = np.hstack([0, np.cumsum(increments)])
            _start_merge_at = start_merge_at + offsets[:-1]
            _end_merge_at = end_merge_at + offsets[1:]
            if np.min(_end_merge_at - _start_merge_at) < 2:
                raise ValueError("You are merging nothing!")
            # prepare for peaklets including S0s
            __merged = np.hstack([_merged, np.full(increments.sum(), True)])
            _peaklets = np.hstack([peaklets, np.hstack(merged_s0s)])
            argsort = strax.stable_argsort(_peaklets["time"])
            merged_s2s = self.merge_peaklets(
                _peaklets[argsort],
                _start_merge_at,
                _end_merge_at,
                __merged[argsort],
                max_buffer,
                max_unmerged=self.unmerged_thresholds,
            )
        else:
            merged_s2s = self.merge_peaklets(
                peaklets,
                start_merge_at,
                end_merge_at,
                _merged,
                max_buffer,
                max_unmerged=self.unmerged_thresholds,
            )

        if (
            len(merged_s2s)
            and np.max((strax.endtime(merged_s2s) - merged_s2s["time"])) > self.max_duration
        ):
            raise ValueError("Merged S2 is too long")

        if self.merge_lone_hits:
            # Updated time and length of lone_hits and sort again:
            # this is an OverlapWindowPlugin, some lone_hits will be reused in the next iteration
            # so do not overwirte the lone_hits
            lh = np.copy(lone_hits)
            del lone_hits
            lh_time_shift = (lh["left"] - lh["left_integration"]) * lh["dt"]
            lh["time"] = lh["time"] - lh_time_shift
            lh["length"] = lh["right_integration"] - lh["left_integration"]
            lh = strax.sort_by_time(lh)

            _store_data_top = "data_top" in self.dtype_for("merged_s2s").names
            _store_data_start = "data_start" in self.dtype_for("merged_s2s").names
            strax.add_lone_hits(
                merged_s2s,
                lh,
                self.to_pe,
                n_top_channels=self.n_top_pmts,
                store_data_top=_store_data_top,
                store_data_start=_store_data_start,
            )

        strax.compute_properties(merged_s2s, n_top_channels=self.n_top_pmts)

        # remove position fields
        merged_s2s = drop_data_field(
            merged_s2s, self.dtype_for("merged_s2s"), "_drop_data_field_merged_s2s"
        )

        # make sure merged has same length as peaklets
        merged = np.zeros(len(is_s2), dtype=bool)
        merged[is_s2] = _merged

        return merged_s2s, merged

    @staticmethod
    def get_left_right(peaklet):
        """Get the left and right boundaries of the peaklet."""
        # The gap is defined as the 90% to 10% area decile distance of the adjacent peaks
        left = (peaklet["area_decile_from_midpoint"][1] + peaklet["median_time"]).astype(int)
        right = (peaklet["area_decile_from_midpoint"][9] + peaklet["median_time"]).astype(int)
        return left, right

    @staticmethod
    def get_gap(y, x):
        """Get the gap between two peaklets."""
        y_left, y_right = MergedS2s.get_left_right(y)
        x_left, x_right = MergedS2s.get_left_right(x)
        dt = x["time"] - y["time"]
        boundaries = np.sort([y_left, y_right, x_left + dt, x_right + dt])
        this_gap = boundaries[2] - boundaries[1]
        return this_gap

    @staticmethod
    @numba.njit(cache=True, nogil=True)
    def get_duration(y, x):
        """Get the duration of the merged peaklets."""
        return max(y["endtime"], x["endtime"]) - min(y["time"], x["time"])

    @staticmethod
    @numba.njit(cache=True, nogil=True)
    def decile_interp(
        x,
        y,
        panel=np.linspace(0, 1, 11),
    ):
        area = x["area"] + y["area"]
        xf = x["area"] / area
        yf = y["area"] / area
        xp = np.concatenate((panel * xf, panel * yf + xf))
        yp = np.concatenate(
            (
                (x["area_decile_from_midpoint"] + x["median_time"]),
                (y["time"] - x["time"]) + (y["area_decile_from_midpoint"] + y["median_time"]),
            )
        )
        decile = np.interp(panel, xp, yp)
        return decile[5], decile - decile[5]

    @staticmethod
    def merge_two_peaks(x, y):
        """Merge two peaklets in order."""
        z = np.array(0, dtype=x.dtype)
        z["time"] = x["time"]
        z["endtime"] = y["endtime"]
        z["area"] = x["area"] + y["area"]
        z["median_time"], z["area_decile_from_midpoint"] = MergedS2s.decile_interp(x, y)
        return z

    @staticmethod
    def get_merge_instructions(
        _peaks,
        start,
        end,
        max_gap,
        max_duration,
        sigma,
        rough_seg,
        sigma_seg,
        rough_mu_bins,
        poisson_max_mu,
        poisson_survival_ratio,
        normal_panel,
        factorial_panel,
        sigma_panel,
        maxexp,
        n_top_pmts,
        p_thresholds,
        dr_thresholds,
        posrec_algo,
        bayesian=True,
        sparse_xy=True,
        uncertainty_weights=True,
        p_value_prioritized=False,
        gap_thresholds=None,
        diagnosing=False,
        disable=True,
    ):
        """Find the group of peaklets to merge.

        There are two ways to merge peaklets:
        1. Bayesian merging: merge peaklets based on the p-value of the time-density merging
        2. Normal merging: merge peaklets based on the gap between the peaklets

        """

        # (x, y) positions of the peaklets
        positions = np.vstack([_peaks[f"x_{posrec_algo}"], _peaks[f"y_{posrec_algo}"]]).T
        if uncertainty_weights:
            contour_area = _peaks[f"position_contour_area_{posrec_algo}"]
        # weights of the peaklets when calculating the weighted mean deviation in (x, y)
        area = _peaks["area"]
        area_top = area * _peaks["area_fraction_top"]

        peaks = np.zeros(len(_peaks), dtype=MergedS2s.copied_dtype)
        for name in peaks.dtype.names:
            if name == "endtime":
                peaks[name] = strax.endtime(_peaks)
            else:
                peaks[name] = _peaks[name]
        peaks_copy = np.copy(peaks)
        peaks_copy.flags.writeable = False

        n_peaks = len(peaks)
        if n_peaks == 0:
            raise ValueError("No peaklets to merge")

        start_index = np.arange(n_peaks)
        # exclusive end index
        end_index = np.arange(n_peaks) + 1
        # keep a copy of the original indices
        start_index_copy = np.copy(start_index)
        start_index_copy.flags.writeable = False
        end_index_copy = np.copy(end_index)
        end_index_copy.flags.writeable = False

        # mask to help keep track of the peaklets that have been examined
        unexamined = np.full(n_peaks, True)
        # mask to help keep track of the peaklets that should not be merged because
        # of too high standard deviation from the main cluster in (x, y) of the peaklets
        merged = np.full(n_peaks, True)

        # approximation of the integration boundaries
        core_bounds = (peaks["time"][1:] + peaks["endtime"][:-1]) // 2
        # here the constraint on boundaries is also to make sure get_window_size covers the gaps
        left_bounds = np.maximum(np.hstack([start, core_bounds]), peaks["time"] - int(max_gap / 2))
        right_bounds = np.minimum(
            np.hstack([core_bounds, end]), peaks["endtime"] + int(max_gap / 2)
        )
        # keep a copy of the original boundaries
        left_bounds_copy = np.copy(left_bounds)
        left_bounds_copy.flags.writeable = False
        right_bounds_copy = np.copy(right_bounds)
        right_bounds_copy.flags.writeable = False

        if diagnosing:
            merged_area = []
            p_values = []
            dr_avgs = []

        def gap_small_enough(y, x):
            log_area = np.log10(y["area"] + x["area"])
            this_gap = MergedS2s.get_gap(y, x)
            gap_threshold = thresholds_interpolation(log_area, gap_thresholds)
            # gap of 90-10% area decile distance should not be larger than the threshold
            return this_gap < gap_threshold

        def can_merge_left(i, peaks):
            # whether peaklet i (from _peak) can be merged to its left unexamined peaklet
            flag = (
                i - 1 >= 0
                and unexamined[i - 1]
                and MergedS2s.get_gap(peaks[i], peaks[i - 1]) < max_gap
                and MergedS2s.get_duration(peaks[i], peaks[i - 1]) <= max_duration
            )
            if not flag:
                return False
            return gap_small_enough(peaks[i], peaks[i - 1])

        def can_merge_right(i, peaks):
            # whether peaklet i (from _peak) can be merged to its right unexamined peaklet
            flag = (
                i < n_peaks
                and unexamined[i]
                and MergedS2s.get_gap(peaks[i - 1], peaks[i]) < max_gap
                and MergedS2s.get_duration(peaks[i - 1], peaks[i]) <= max_duration
            )
            if not flag:
                return False
            return gap_small_enough(peaks[i - 1], peaks[i])

        argsort = strax.stable_argsort(peaks["area"])
        for i in tqdm(argsort[::-1], disable=disable):
            if not unexamined[i]:
                continue
            p = np.array([1.0, 1.0])
            p_threshold = np.array([0.0, 0.0])
            left = True
            dr_avg = 0.0
            # in the while loops, the peaklets will be merged until the p-value
            # is smaller than the threshold or the peaklets can not be merged anymore
            # i will NOT be updated in the while loop
            while np.any(p >= p_threshold) and unexamined[i]:
                indices = []
                # please mind here that the definition of gaps should
                # be consistent with in the merging algorithm
                # and the merged peak should not be longer than the max duration
                if can_merge_left(start_index[i], peaks):
                    indices.append(start_index[i] - 1)
                else:
                    indices.append(None)
                if can_merge_right(end_index[i], peaks):
                    indices.append(end_index[i])
                else:
                    indices.append(None)
                p = []
                p_threshold = []
                _area = []
                # get p-values
                for j in indices:
                    if j is None:
                        p.append(-1.0)
                        p_threshold.append(1.0)
                        _area.append(-1.0)
                        continue
                    # if area is non-positive, "merge" in time
                    # but still skip it later in (x, y) because its weight is nan
                    if not peaks["area"][j] > 0:
                        p.append(1.0)
                        p_threshold.append(-1.0)
                        _area.append(peaks["area"][j])
                        continue
                    log_area = np.log10(peaks["area"][i] + peaks["area"][j])
                    if bayesian:
                        p_ = get_p_value(
                            peaks[i],
                            peaks[j],
                            left_bounds[i],
                            right_bounds[i],
                            left_bounds[j],
                            right_bounds[j],
                            sigma,
                            rough_seg,
                            sigma_seg,
                            rough_mu_bins,
                            poisson_max_mu,
                            poisson_survival_ratio,
                            normal_panel,
                            factorial_panel,
                            sigma_panel,
                            maxexp,
                        )
                        p_threshold_ = thresholds_interpolation(
                            log_area,
                            p_thresholds,
                        )
                    else:
                        # this is kept for diagnosing and merged_s2s_he
                        p_ = 1.0
                        p_threshold_ = 0.0
                    p.append(p_)
                    p_threshold.append(p_threshold_)
                    _area.append(peaks["area"][j])
                p = np.array(p)
                p_threshold = np.array(p_threshold)
                _area = np.array(_area)
                examined = slice(start_index[i], end_index[i])

                if diagnosing:
                    merged_area.append(area[examined][merged[examined]].sum())

                if np.all(p < p_threshold):
                    # this will not allow merging of the already examined peaklets
                    unexamined[examined] = False
                else:
                    # whether test the proposal with larger area first
                    if p_value_prioritized:
                        left = p[0] > p[1]
                    else:
                        left = _area[0] > _area[1]
                    # if test left first
                    if left:
                        if p[0] >= p_threshold[0]:
                            left = True
                        else:
                            assert p[1] >= p_threshold[1]
                            left = False
                    else:
                        if p[1] >= p_threshold[1]:
                            left = False
                        else:
                            assert p[0] >= p_threshold[0]
                            left = True
                    # slice indicating the direction of merging
                    if left:
                        direction = [indices[0], indices[0] + 1]
                        index = indices[0]
                    else:
                        direction = [indices[1] - 1, indices[1]]
                        index = indices[1]
                    # slice indicating the peaklets to be merged
                    start_idx = start_index[direction[0]]
                    end_idx = end_index[direction[1]]
                    merging = slice(start_idx, end_idx)

                    if sparse_xy:
                        # calculate weighted averaged deviation of peaklets from the main cluster
                        if uncertainty_weights:
                            weights = 1 / contour_area[merging][merged[merging]]
                        else:
                            weights = area_top[merging][merged[merging]]

                        dr_avg = weighted_averaged_dr(
                            positions[merging, 0][merged[merging]],
                            positions[merging, 1][merged[merging]],
                            weights,
                        )
                        # do we really merge the peaklets?
                        dr_threshold_ = thresholds_interpolation(
                            np.log10(area_top[merging][merged[merging]].sum()),
                            dr_thresholds,
                        )
                        merge = dr_avg < dr_threshold_
                    else:
                        merge = True

                    if merge:
                        merged_peak = MergedS2s.merge_two_peaks(
                            peaks[direction[0]], peaks[direction[1]]
                        )
                        # check if the merged peak is too long
                        if merged_peak["endtime"] - merged_peak["time"] > max_duration:
                            raise ValueError("Merged S2 is too long")
                    else:
                        merged[index] = False
                        merged_peak = peaks[i]

                    # update merging peaks and boundaries
                    # update ALL peaks in the slice, this is only temporarily needed
                    # because we can not foresee in which order the peaks will be merged
                    peaks[merging] = merged_peak
                    start_index[merging] = start_index[merging.start]
                    end_index[merging] = end_index[merging.stop - 1]
                    left_bounds[merging] = left_bounds[merging.start]
                    right_bounds[merging] = right_bounds[merging.stop - 1]

                if diagnosing:
                    if left:
                        p_values.append(p[0])
                    else:
                        p_values.append(p[1])
                    if sparse_xy:
                        dr_avgs.append(dr_avg)
                    else:
                        dr_avgs.append(0.0)

        if np.any(np.diff(start_index) < 0) or np.any(np.diff(end_index) < 0):
            raise ValueError("Indices are not sorted!")
        n_peaklets = end_index - start_index
        need_merging = n_peaklets > 1
        start_index = np.unique(start_index[need_merging])
        end_index = np.unique(end_index[need_merging])
        if diagnosing:
            return start_index, end_index, merged, merged_area, p_values, dr_avgs
        return start_index, end_index, merged

    @staticmethod
    def merge_peaklets(
        peaklets,
        start_merge_at,
        end_merge_at,
        merged,
        max_buffer=int(1e5),
        max_unmerged=None,
        merged_all=False,
    ):
        if merged_all:
            # execute earlier to prevent peaklets from being overwritten
            # mark the peaklets can be merged by time-density but not
            # by position-density as type FAR_XYPOS_S2_TYPE
            _merged_s2s = np.copy(peaklets[~merged])
            _merged_s2s["type"] = FAR_XYPOS_S2_TYPE
        # mark the peaklets can be merged by time-density and position-density as type 2
        merged_s2s = strax.merge_peaks(
            peaklets,
            start_merge_at,
            end_merge_at,
            merged=merged,
            max_buffer=max_buffer,
        )
        merged_s2s["type"] = 2

        # if the number of type 20 peaklets inside a peak is larger than the threshold
        # or the area of type 20 peaklets inside a peak is larger than the threshold
        # mark the peaklets as type WIDE_XYPOS_S2_TYPE
        if max_unmerged is not None:
            n_de = []
            area = []
            area_de = []
            for i in range(len(merged_s2s)):
                sl = slice(start_merge_at[i], end_merge_at[i])
                n_de.append(np.sum(~merged[sl]))
                area.append(peaklets["area"][sl][merged[sl]].sum())
                area_de.append(peaklets["area"][sl][~merged[sl]].sum())
            n_de = np.array(n_de)
            area = np.array(area)
            area_de = np.array(area_de)
            # if the number of type 20 peaklets inside a peak is larger than the threshold
            mask = n_de > max_unmerged[0]
            # if the area of type 20 peaklets inside a peak is larger than the (both) threshold(s)
            mask |= (area_de > max_unmerged[1]) & (area_de > area * max_unmerged[2])
            merged_s2s["type"] = np.where(mask, WIDE_XYPOS_S2_TYPE, merged_s2s["type"])

        if merged_all:
            merged_s2s = strax.sort_by_time(np.concatenate([_merged_s2s, merged_s2s]))
        return merged_s2s


@numba.njit(cache=True, nogil=True)
def get_p_value(
    y,
    x,
    ly,
    ry,
    lx,
    rx,
    sigma,
    rough_seg,
    sigma_seg,
    rough_mu_bins,
    poisson_max_mu,
    poisson_survival_ratio,
    normal_panel,
    factorial_panel,
    sigma_panel,
    maxexp,
):
    # better constraint the mu in boundaries of waveform
    # because the number of bins is not high enough comprising for computation efficiency
    mu = rough_mu(y, x, rough_mu_bins)

    t0 = y["time"]
    y_integrated = normal_cdf(((ry - t0) - mu[:, None]) / sigma, normal_panel)
    y_integrated -= normal_cdf(((ly - t0) - mu[:, None]) / sigma, normal_panel)
    x_integrated = normal_cdf(((rx - t0) - mu[:, None]) / sigma, normal_panel)
    x_integrated -= normal_cdf(((lx - t0) - mu[:, None]) / sigma, normal_panel)
    # keep numerical stability
    eps = 2.220446049250313e-16  # np.finfo(float).eps
    y_integrated += eps
    x_integrated += eps

    pmf = get_posterior(
        mu,
        sigma,
        y["area"],
        y["area_decile_from_midpoint"],
        y_integrated,
        y["time"],
        y["median_time"],
        x["time"],
        x["median_time"],
        rough_seg,
        maxexp,
    ).flatten()

    non_zero = pmf > 0
    if not np.any(non_zero):
        return 0.0
    # the mask non_zero here is also implicit flatten
    ps = p_values(
        sigma_panel[non_zero],
        y["area"],
        y_integrated.flatten()[non_zero],
        x["area"],
        x_integrated.flatten()[non_zero],
        rough_seg,
        sigma_seg,
        poisson_max_mu,
        poisson_survival_ratio,
        normal_panel,
        factorial_panel,
    )

    p = np.sum(ps * pmf[non_zero])
    if np.isnan(p):
        raise RuntimeError("p-value is NaN")
    return p


@numba.njit(cache=True, nogil=True)
def thresholds_interpolation(log_area, thresholds):
    """Return threshold for log_area of the merged S2 with linear interpolation given the points in
    thresholds.

    :param log_area: Log 10 area of the merged S2
    :param thresholds: tuple (n, 2) of fix points for interpolation.

    """
    if log_area < thresholds[0, 0]:
        return thresholds[1, 0]
    if log_area > thresholds[0, -1]:
        return thresholds[1, -1]
    return np.interp(log_area, thresholds[0], thresholds[1])


@numba.njit(cache=True, nogil=True)
def rough_mu(y, x, rough_mu_bins):
    """Get rough bins for mu as the integration space."""
    mu = np.linspace(
        min(y["time"], x["time"]) - y["time"],
        max(y["endtime"], x["endtime"]) - y["time"],
        rough_mu_bins,
    )
    return mu


@numba.njit(cache=True, nogil=True)
def poisson_pmf(k, mu, factorial_panel):
    """Probability mass function of Poisson distribution, numba decorated."""
    return mu**k / np.exp(mu) / factorial_panel[k]


@numba.njit(cache=True, nogil=True)
def normal_cdf(x, normal_panel):
    """Cumulative density function of standard normal distribution, numba decorated."""
    return np.interp(x, normal_panel[0], normal_panel[1])


@numba.njit(cache=True, nogil=True)
def get_posterior(
    mu,
    sigma,
    y_area,
    y_area_decile_from_midpoint,
    y_integrated,
    y_time,
    y_median_time,
    x_time,
    x_median_time,
    rough_seg,
    maxexp,
):
    """Get the posterior of (mu, sigma) based on the waveform y."""
    # production of normal PDF
    y_t = y_area_decile_from_midpoint + y_median_time
    y_t = (y_t[:-1] + y_t[1:]) / 2
    y_nse = y_area / rough_seg
    sigma2 = 2 * sigma**2
    log_sigma = np.log(sigma)
    # dt is median between median time of two peaklets
    dt = np.abs((y_time - x_time) + (y_median_time - x_median_time))
    # because the number of electrons are uniformly distributed in y_t
    log_pdf = -(((y_t - mu[:, None]) ** 2).sum(axis=1) * (y_nse / 10))[:, None] / sigma2
    # complete normal distribution PDF log_sigma * y_nse
    # add log prior dt**2 / 4 / sigma2 + 2 * log_sigma
    # account for the bounded sampling of y np.log(y_integrated) * y_nse
    log_pdf -= np.log(y_integrated) * y_nse + log_sigma * (y_nse + 2) + dt**2 / 4 / sigma2
    # to prevent numerical overflow
    log_pdf -= log_pdf.max()
    log_pdf += maxexp
    pdf = np.exp(log_pdf)
    pmf = pdf / pdf.sum()
    if np.any(np.isnan(pmf)):
        raise RuntimeError("pmf is NaN")
    return pmf


@numba.njit(cache=True, nogil=True)
def p_values(
    sigma,
    y_area,
    y_integrated,
    x_area,
    x_integrated,
    rough_seg,
    sigma_seg,
    poisson_max_mu,
    poisson_survival_ratio,
    normal_panel,
    factorial_panel,
):
    """The p-value if observe the x_area given the y_area, numba decorated."""
    # distribution of allowed area
    expected_nse = x_integrated * y_area / y_integrated / rough_seg
    ps = []
    for mu in expected_nse:
        # the number of electron follows Poisson distribution
        if mu > poisson_max_mu:
            ps.append(1 - normal_cdf((x_area / rough_seg - mu) / np.sqrt(mu), normal_panel))
        else:
            cdf = poisson_pmf(0, mu, factorial_panel)
            k = 1
            p = 0
            while cdf < 1 - poisson_survival_ratio:
                pmf = poisson_pmf(k, mu, factorial_panel)
                cdf += pmf
                p += pmf * (
                    1 - normal_cdf((x_area - k * rough_seg) / (k**0.5 * sigma_seg), normal_panel)
                )
                k += 1
            ps.append(p)
    ps = np.array(ps)
    return ps


@numba.njit(cache=True, nogil=True)
def weighted_averaged_dr(x, y, weights):
    """Weighted average deviation from weighted average (x, y)"""
    mask = weights > 0
    mask &= ~np.isnan(x)
    mask &= ~np.isnan(x)
    # do not merge any S2 looks weird
    if not np.all(mask):
        return np.nan
    x_avg = np.average(x[mask], weights=weights[mask])
    y_avg = np.average(y[mask], weights=weights[mask])
    dr = np.sqrt((x - x_avg) ** 2 + (y - y_avg) ** 2)
    dr_avg = np.average(dr[mask], weights=weights[mask])
    return dr_avg



from typing import Tuple

import numpy as np
import strax
import straxen
from straxen.common import rotate_perp_wires
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO

export, __all__ = strax.exporter()


@export
class CorrectedAreas(strax.Plugin):
    """Plugin which applies light collection efficiency maps and electron life time to the data.

    Computes the cS1/cS2 for the main/alternative S1/S2 as well as the
    corrected life time.
    Note:
        Please be aware that for both, the main and alternative S1, the
        area is corrected according to the xy-position of the main S2.
        There are now 3 components of cS2s: cs2_top, cS2_bottom and cs2.
        cs2_top and cs2_bottom are corrected by the corresponding maps,
        and cs2 is the sum of the two.

    """

    __version__ = "0.5.3"

    depends_on: Tuple[str, ...] = ("event_basics", "event_positions")

    # Descriptor configs
    elife = straxen.URLConfig(
        default="xedocs://electron_lifetimes?attr=value&run_id=plugin.run_id&version=ONLINE",
        help="electron lifetime in [ns]",
    )

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )
    s1_xyz_map = straxen.URLConfig(
        default="xedocs://s1_xyz_maps"
        "?run_id=plugin.run_id"
        "&algorithm=plugin.default_reconstruction_algorithm&attr=map&version=ONLINE",
        cache=True,
    )
    s2_xy_map = straxen.URLConfig(
        default="xedocs://s2_xy_maps"
        "?run_id=plugin.run_id"
        "&algorithm=plugin.default_reconstruction_algorithm&attr=map&version=ONLINE",
        cache=True,
    )

    # average SE gain for a given time period. default to the value of this run in ONLINE model
    # thus, by default, there will be no time-dependent correction according to se gain
    avg_se_gain = straxen.URLConfig(
        default="xedocs://avg_se_gains?run_id=plugin.run_id&version=ONLINE&attr=value",
        help="Nominal single electron (SE) gain in PE / electron extracted. "
        "Data will be corrected to this value",
    )

    # se gain for this run, allowing for using xedocs. default to online
    se_gain = straxen.URLConfig(
        default="take://objects-to-dict://"
        "xedocs://se_gains"
        "?partition=all_tpc&run_id=plugin.run_id&sort=partition"
        "&as_list=True&key_attr=partition&value_attr=value&take=all_tpc&version=ONLINE",
        help="Actual SE gain for a given run (allows for time dependence)",
    )

    # relative extraction efficiency which can change with time and modeled by xedocs.
    rel_extraction_eff = straxen.URLConfig(
        default="take://objects-to-dict://"
        "xedocs://rel_extraction_effs"
        "?partition=all_tpc&run_id=plugin.run_id&sort=partition"
        "&as_list=True&key_attr=partition&value_attr=value&take=all_tpc&version=ONLINE",
        help="Relative extraction efficiency for this run (allows for time dependence)",
    )

    # relative light yield
    # defaults to no correction
    rel_light_yield = straxen.URLConfig(
        default="xedocs://relative_light_yield?attr=value&run_id=plugin.run_id&version=ONLINE",
        help="Relative light yield (allows for time dependence)",
    )

    # Single electron gain partition
    # AB and CD partitons distiguished based on
    # linear and circular regions
    # SR0 values set as default
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=jlong:sr0_2_region_se_correction
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:noahhood:corrections:se_gain_ee_final
    single_electron_gain_partition = straxen.URLConfig(
        default={"linear": 28, "circular": 60},
        help=(
            "Two distinct patterns of evolution of single electron corrections between AB and CD. "
            "Distinguish thanks to linear and circular regions"
        ),
    )

    # cS2 AFT correction due to photon ionization
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:zihao:sr1_s2aft_photonionization_correction
    cs2_bottom_top_ratio_correction = straxen.URLConfig(
        default=1, help="Scaling factor for cS2 AFT correction due to photon ionization"
    )

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields

        for peak_type, peak_name in zip(["", "alt_"], ["main", "alternate"]):
            # Only apply
            dtype += [
                (f"{peak_type}cs1", np.float32, f"Corrected area of {peak_name} S1 [PE]"),
                (
                    f"{peak_type}cs1_wo_timecorr",
                    np.float32,
                    f"Corrected area of {peak_name} S1 (before LY correction) [PE]",
                ),
            ]
            names = ["_wo_timecorr", "_wo_picorr", "_wo_elifecorr", ""]
            descriptions = ["S2 xy", "SEG/EE", "photon ionization", "elife"]
            for i, name in enumerate(names):
                if i == len(names) - 1:
                    description = ""
                elif i == 0:
                    # special treatment for wo_timecorr, apply elife correction
                    description = " (before " + " + ".join(descriptions[i + 1 : -1])
                    description += (
                        ", after " + " + ".join(descriptions[: i + 1] + descriptions[-1:]) + ")"
                    )
                else:
                    description = " (before " + " + ".join(descriptions[i + 1 :])
                    description += ", after " + " + ".join(descriptions[: i + 1]) + ")"
                dtype += [
                    (
                        f"{peak_type}cs2{name}",
                        np.float32,
                        f"Corrected area of {peak_name} S2{description} [PE]",
                    ),
                    (
                        f"{peak_type}cs2_area_fraction_top{name}",
                        np.float32,
                        (
                            "Fraction of area seen by the top PMT array for corrected "
                            f"{peak_name} S2{description}"
                        ),
                    ),
                ]
        return dtype

    def ab_region(self, x, y):
        new_x, new_y = rotate_perp_wires(x, y)
        cond = new_x < self.single_electron_gain_partition["linear"]
        cond &= new_x > -self.single_electron_gain_partition["linear"]
        cond &= new_x**2 + new_y**2 < self.single_electron_gain_partition["circular"] ** 2
        return cond

    def cd_region(self, x, y):
        return ~self.ab_region(x, y)

    def s2_map_names(self):
        # S2 top and bottom are corrected separately, and cS2 total is the sum of the two
        # figure out the map name
        if len(self.s2_xy_map.map_names) > 1:
            s2_top_map_name = "map_top"
            s2_bottom_map_name = "map_bottom"
        else:
            s2_top_map_name = "map"
            s2_bottom_map_name = "map"

        return s2_top_map_name, s2_bottom_map_name

    def seg_ee_correction_preparation(self):
        """Get single electron gain and extraction efficiency options."""
        self.regions = {"ab": self.ab_region, "cd": self.cd_region}

        # setup SEG and EE corrections
        # if they are dicts, we just leave them as is
        # if they are not, we assume they are floats and
        # create a dict with the same correction in each region
        if isinstance(self.se_gain, dict):
            seg = self.se_gain
        else:
            seg = {key: self.se_gain for key in self.regions}

        if isinstance(self.avg_se_gain, dict):
            avg_seg = self.avg_se_gain
        else:
            avg_seg = {key: self.avg_se_gain for key in self.regions}

        if isinstance(self.rel_extraction_eff, dict):
            ee = self.rel_extraction_eff
        else:
            ee = {key: self.rel_extraction_eff for key in self.regions}

        return seg, avg_seg, ee

    def compute(self, events):
        result = np.zeros(len(events), self.dtype)
        result["time"] = events["time"]
        result["endtime"] = events["endtime"]

        # S1 corrections depend on the actual corrected event position.
        # We use this also for the alternate S1; for e.g. Kr this is
        # fine as the S1 correction varies slowly.
        event_positions = np.vstack([events["x"], events["y"], events["z"]]).T

        for peak_type in ["", "alt_"]:
            result[f"{peak_type}cs1_wo_timecorr"] = events[f"{peak_type}s1_area"] / self.s1_xyz_map(
                event_positions
            )
            result[f"{peak_type}cs1"] = result[f"{peak_type}cs1_wo_timecorr"] / self.rel_light_yield

        # S2 corrections
        s2_top_map_name, s2_bottom_map_name = self.s2_map_names()
        seg, avg_seg, ee = self.seg_ee_correction_preparation()

        # now can start doing corrections
        for peak_type in ["", "alt_"]:
            # S2(x,y) corrections use the observed S2 positions
            s2_positions = np.vstack([events[f"{peak_type}s2_x"], events[f"{peak_type}s2_y"]]).T

            # corrected S2 with S2(x,y) map only, i.e. no elife correction
            # this is for S2-only events which don't have drift time info
            s2_xy_top = self.s2_xy_map(s2_positions, map_name=s2_top_map_name)
            cs2_top_xycorr = (
                events[f"{peak_type}s2_area"]
                * events[f"{peak_type}s2_area_fraction_top"]
                / s2_xy_top
            )
            s2_xy_bottom = self.s2_xy_map(s2_positions, map_name=s2_bottom_map_name)
            cs2_bottom_xycorr = (
                events[f"{peak_type}s2_area"]
                * (1 - events[f"{peak_type}s2_area_fraction_top"])
                / s2_xy_bottom
            )

            # collect electron lifetime correction
            # for electron lifetime corrections to the S2s,
            # use drift time computed using the main S1.
            el_string = peak_type + "s2_interaction_" if peak_type == "alt_" else peak_type
            elife_correction = np.exp(events[f"{el_string}drift_time"] / self.elife)

            # collect SEG and EE corrections
            seg_ee_corr = np.zeros(len(events))
            for partition, func in self.regions.items():
                # partitioned SEG and EE
                partition_mask = func(events[f"{peak_type}s2_x"], events[f"{peak_type}s2_y"])
                # correct for SEG and EE
                seg_ee_corr[partition_mask] = seg[partition] / avg_seg[partition] * ee[partition]

            # apply S2 xy correction
            cs2_xycorr = cs2_top_xycorr + cs2_bottom_xycorr
            result[f"{peak_type}cs2_wo_timecorr"] = cs2_xycorr * elife_correction
            result[f"{peak_type}cs2_area_fraction_top_wo_timecorr"] = cs2_top_xycorr / cs2_xycorr

            # apply SEG and EE correction
            cs2_top_wo_picorr = cs2_top_xycorr / seg_ee_corr
            cs2_bottom_wo_picorr = cs2_bottom_xycorr / seg_ee_corr
            cs2_wo_picorr = cs2_top_wo_picorr + cs2_bottom_wo_picorr
            result[f"{peak_type}cs2_wo_picorr"] = cs2_wo_picorr
            result[f"{peak_type}cs2_area_fraction_top_wo_picorr"] = (
                cs2_top_wo_picorr / result[f"{peak_type}cs2_wo_picorr"]
            )

            # apply photon ionization intensity and cS2 AFT correction (see #1247)
            # cS2 bottom should be corrected by photon ionization, but not cS2 top
            cs2_top_wo_elifecorr = cs2_top_wo_picorr
            cs2_bottom_wo_elifecorr = cs2_bottom_wo_picorr * self.cs2_bottom_top_ratio_correction
            cs2_wo_elifecorr = cs2_top_wo_elifecorr + cs2_bottom_wo_elifecorr
            # scale top and bottom to ensure total cS2 is conserved, since the time
            # dependence of it has been already corrected by SEG correction
            cs2_top_wo_elifecorr *= cs2_wo_picorr / cs2_wo_elifecorr
            cs2_bottom_wo_elifecorr *= cs2_wo_picorr / cs2_wo_elifecorr
            cs2_wo_elifecorr = cs2_wo_picorr
            result[f"{peak_type}cs2_wo_elifecorr"] = cs2_wo_elifecorr
            result[f"{peak_type}cs2_area_fraction_top_wo_elifecorr"] = (
                cs2_top_wo_elifecorr / result[f"{peak_type}cs2_wo_elifecorr"]
            )

            # apply electron lifetime correction
            result[f"{peak_type}cs2"] = result[f"{peak_type}cs2_wo_elifecorr"] * elife_correction
            result[f"{peak_type}cs2_area_fraction_top"] = result[
                f"{peak_type}cs2_area_fraction_top_wo_elifecorr"
            ]
        return result



import strax
from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.peaks.peaks_vanilla import PeaksVanilla

export, __all__ = strax.exporter()


@export
class PeaksHighEnergy(PeaksVanilla):
    __doc__ = HE_PREAMBLE + (PeaksVanilla.__doc__ or "")

    __version__ = "0.0.1"
    depends_on = ("peaklets_he", "peaklet_classification_he", "merged_s2s_he")
    data_kind = "peaks_he"
    provides = "peaks_he"
    child_ends_with = "_he"

    def infer_dtype(self):
        return self.deps["peaklets_he"].dtype_for("peaklets")

    def compute(self, peaklets_he, merged_s2s_he):
        return super().compute(peaklets_he, merged_s2s_he)



import numpy as np
import numba
import strax
import straxen


export, __all__ = strax.exporter()


@export
class PeakProximity(strax.OverlapWindowPlugin):
    """Look for peaks around a peak to determine how many peaks are in proximity (in time) of a
    peak."""

    __version__ = "0.5.0"

    depends_on = "peak_basics"
    dtype = [
        ("n_competing", np.int32, "Number of nearby larger or slightly smaller peaks"),
        (
            "n_competing_left",
            np.int32,
            "Number of larger or slightly smaller peaks left of the main peak",
        ),
        (
            "t_to_prev_peak",
            np.int64,
            "Time between end of previous peak and start of this peak [ns]",
        ),
        ("t_to_next_peak", np.int64, "Time between end of this peak and start of next peak [ns]"),
        ("t_to_nearest_peak", np.int64, "Smaller of t_to_prev_peak and t_to_next_peak [ns]"),
    ] + strax.time_fields

    min_area_fraction = straxen.URLConfig(
        default=0.5,
        infer_type=False,
        help=(
            "The area of competing peaks must be at least "
            "this fraction of that of the considered peak"
        ),
    )

    nearby_window = straxen.URLConfig(
        default=int(1e7),
        infer_type=False,
        help="Peaks starting within this time window (on either side) in ns count as nearby.",
    )

    peak_max_proximity_time = straxen.URLConfig(
        default=int(1e8),
        infer_type=False,
        help="Maximum value for proximity values such as t_to_next_peak [ns]",
    )

    def get_window_size(self):
        return self.peak_max_proximity_time

    def compute(self, peaks):
        # only consider S1 and S2 peaks
        mask = np.isin(peaks["type"], [1, 2])
        windows = strax.touching_windows(peaks[mask], peaks, window=self.nearby_window)
        n_left, n_tot = self.find_n_competing(
            peaks[mask], peaks, windows, fraction=self.min_area_fraction
        )

        t_to_prev_peak = np.ones(len(peaks), dtype=np.int64) * self.peak_max_proximity_time
        t_to_prev_peak[1:] = peaks["time"][1:] - peaks["endtime"][:-1]

        t_to_next_peak = t_to_prev_peak.copy()
        t_to_next_peak[:-1] = peaks["time"][1:] - peaks["endtime"][:-1]

        return dict(
            time=peaks["time"],
            endtime=strax.endtime(peaks),
            n_competing=n_tot,
            n_competing_left=n_left,
            t_to_prev_peak=t_to_prev_peak,
            t_to_next_peak=t_to_next_peak,
            t_to_nearest_peak=np.minimum(t_to_prev_peak, t_to_next_peak),
        )

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def find_n_competing(peaks_in_roi, peaks, windows, fraction):
        n_left = np.zeros(len(peaks), dtype=np.int32)
        n_tot = n_left.copy()
        areas_in_roi = peaks_in_roi["area"]
        areas = peaks["area"]

        dig = np.searchsorted(peaks_in_roi["center_time"], peaks["center_time"])
        for i, peak in enumerate(peaks):
            left_i, right_i = windows[i]
            threshold = areas[i] * fraction
            n_left[i] = np.sum(areas_in_roi[left_i : dig[i]] > threshold)
            n_tot[i] = n_left[i] + np.sum(areas_in_roi[dig[i] : right_i] > threshold)

        return n_left, n_tot
